%======================================================================
% MODULE: planner-ME-GR-SP.p
%
% The state-space representation of the simple means-ends planner
% with goal regression.
% 
% See Figure 16.9, p. 420 of
% "Prolog Programming for Artificial Intelligence", 2d Edition
%                        - by Ivan Bratko
%======================================================================

:- op( 300, xfy, ->).

s( Goals -> NextAction, NewGoals -> Action, 1) :-   % All costs are 1
        member( Goal, Goals),
        achieves( Action, Goal),
        can( Action, Condition),
        preserves( Action, Goals),
        regress( Goals, Action, NewGoals).

goal( Goals -> Action) :-
        start( State),                      % User-defined initial situation
        satisfied( State, Goals).           % Goals true in initial situation

h( Goals -> Action, H) :-                   % Heuristic estimate
        start( State),
        delete( Goals, State, Unsatisfied), % Unsatisfied goals
        length( Unsatisfied, H).            % Number of unsatisfied goals

satisfied( State, Goals) :-
        delete( Goals, State, []).      % All Goals in State

achieves( Action, Goal) :-
        add( Action, Goals),
        member( Goal, Goals).

preserves( Action, Goals) :-            % Action does not destroy Goals
        del( Action, Relations),
        not(( member( Goal, Relations), member( Goal, Goals) )).

regress( Goals, Action, RegressedGoals) :-       % Regress Goals through Action
        add( Action, NewRelations),
        delete( Goals, NewRelations, RestGoals),
        can( Action, Condition),
        addnew( Condition, RestGoals, RegressedGoals). % Add precondition,
                                                       %  check impossibility

%----------------------------------------------------------------------
% addnew( NewGoals, OldGoals, AllGoals):
%   AllGoals is the union of NewGoals and OldGoals
%   NewGoals and OldGoals must be compatible
%----------------------------------------------------------------------

addnew( [], L, L).

addnew( [Goal | _], Goals, _) :-
        impossible( Goal, Goals),                % Goal incompatible with Goals
        !,
        fail.                                    % Cannot be added

addnew( [X | L1], L2, L3) :-
        member( X, L2),
        !,
        addnew( L1, L2, L3).                     % Ignore duplicate

addnew( [X | L1], L2, [X | L3]) :-
        addnew( L1, L2, L3).
